/*===============================================================================================
 Copyright (c), Jari Suominen 2008-2012.
===============================================================================================*/
//#include "../../../examples/common/wincompat.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <alsa/asoundlib.h>
#include <fmodex/fmod_event.hpp>
#include <fmodex/fmod_errors.h>
#include "jarse.h"
#include <stdint.h>   /* Standard types */
#include <string.h>   /* String function definitions */
#include <fcntl.h>    /* File control definitions */
#include <errno.h>    /* Error number definitions */
#include <termios.h>  /* POSIX terminal control definitions */
#include <sys/ioctl.h>
#include <getopt.h>


bool variationActive = false;
int fd;

FMOD::MusicSystem *musicsystem;
FMOD::EventSystem *eventsystem;
int midiin = 0;
int midiout = 0;
snd_seq_t *seq_handle;

/*
	Initializing ALSA MIDI.
	It is highly unlikely that you want to edit this or even check what it has eaten.
 */
snd_seq_t *open_seq() {

  snd_seq_t *seq_handle;
  int portid;

  if (snd_seq_open(&seq_handle, "hw", SND_SEQ_OPEN_DUPLEX, 0) < 0) {
    fprintf(stderr, "Error opening ALSA sequencer.\n");
    exit(1);
  }
  snd_seq_set_client_name(seq_handle, "hurja");
  if ((midiin = snd_seq_create_simple_port(seq_handle, "MIDI in",
            SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE,
            SND_SEQ_PORT_TYPE_APPLICATION)) < 0) {
    fprintf(stderr, "Error creating sequencer port.\n");
    exit(1);
  }
  if ((midiout = snd_seq_create_simple_port(seq_handle, "MIDI out",
            SND_SEQ_PORT_CAP_READ|SND_SEQ_PORT_CAP_SUBS_READ,
            SND_SEQ_PORT_TYPE_APPLICATION)) < 0) {
    fprintf(stderr, "Error creating sequencer port.\n");
    exit(1);
  }
  return(seq_handle);
}


void ERRCHECK(FMOD_RESULT result) {
  if (result != FMOD_OK) {
    printf("FMOD error! (%d) %s\n", result, FMOD_ErrorString(result));
    exit(-1);
  }
}

void next() {
  eventsystem->update();
  musicsystem->setParameterValue(MUSICPARAM_JARSE_VARIATION,1);
  musicsystem->setParameterValue(MUSICPARAM_JARSE_NEXT,0);
  printf("NEXT!\n");
  fflush(stdout);
}

void variation() {
  eventsystem->update();
  musicsystem->setParameterValue(MUSICPARAM_JARSE_NEXT,1);
  musicsystem->setParameterValue(MUSICPARAM_JARSE_VARIATION,0);
  printf("VARIATION!\n");
  fflush(stdout);
}

void reset() {
  eventsystem->update();
  musicsystem->setParameterValue(MUSICPARAM_JARSE_NEXT,1);
  musicsystem->setParameterValue(MUSICPARAM_JARSE_VARIATION,1);
  printf("RESET!\n");
  fflush(stdout);
}

FMOD_RESULT resetParams(FMOD_MUSIC_CALLBACKTYPE type, void * p, void * p2, void * ud) {
	snd_seq_event_t ev;
	snd_seq_ev_clear(&ev);
	snd_seq_ev_set_direct(&ev);
	snd_seq_ev_set_source(&ev, midiout);
	snd_seq_ev_set_subs(&ev);

	int operation, channel, param1, param2;

	if (type==FMOD_MUSIC_CALLBACKTYPE_BEAT) {
//		snd_seq_ev_set_noteon(&ev, 10, 11, 11);
		snd_seq_ev_set_noteon(&ev, 6, 0, 0);
		printf("BEAT!\n");
		fflush(stdout);
	}
	else if (type==FMOD_MUSIC_CALLBACKTYPE_SEGMENT_START) {
		float f;
		musicsystem->getParameterValue(MUSICPARAM_JARSE_VARIATION,&f);
		if (f==0) {
			//write_int(51,fd);
		} else {
			//write_int(48,fd);
		}		
		//musicsystem->setParameterValue(MUSICPARAM_JARSE_NEXT,1);
		//musicsystem->setParameterValue(MUSICPARAM_JARSE_VARIATION,1);
		printf("SEGMENT START!\n");
		reset();
		snd_seq_ev_set_noteon(&ev, 10, 11, 12);
		fflush(stdout);
	} else {
		printf("SOMETHING ELSE TRIGGERED!\n");
		snd_seq_ev_set_noteon(&ev, 10, 11, 13);
		fflush(stdout);
	}
	snd_seq_event_output_direct(seq_handle, &ev);
	snd_seq_drain_output(seq_handle);
	return FMOD_OK;
}

/*
	This function processes the MIDI messages arriving from MIDI keyboard through ALSA MIDI.
	Only noteon and noteoff messages are used!
 */
void midi_action(snd_seq_t *seq_handle) {
	snd_seq_event_t *ev;

	do {
		snd_seq_event_input(seq_handle, &ev);
		if (ev->data.control.channel==0) {
			switch (ev->type) {

			case SND_SEQ_EVENT_PGMCHANGE:
				if ((int)ev->data.control.value==1) {
					ERRCHECK(musicsystem->promptCue(MUSICCUE_JARSE_BO));
					ERRCHECK(musicsystem->setPaused(false));
				} else if ((int)ev->data.control.value==2) {
					ERRCHECK(musicsystem->promptCue(MUSICCUE_JARSE_KORPIKUUSEN));
					ERRCHECK(musicsystem->setPaused(false));
				} else if ((int)ev->data.control.value==8) {
					//ERRCHECK(musicsystem->promptCue(MUSICCUE_JARSE_ALAS_2));
					//ERRCHECK(musicsystem->setPaused(false));
					ERRCHECK(musicsystem->setPaused(true));
				} else if ((int)ev->data.control.value==7) {
					ERRCHECK(musicsystem->setPaused(false));
				} else if ((int)ev->data.control.value==5) {
					variation();
				} else if ((int)ev->data.control.value==6) {
					next();
				}
				break;
			}
		}
		else if (ev->data.control.channel==11) {
			switch (ev->type) {
      				case SND_SEQ_EVENT_NOTEON:
					if ((int)ev->data.note.note==2) {
						variation();
					} else if ((int)ev->data.note.note==1) {
						next();
					}
					break;
				case SND_SEQ_EVENT_CONTROLLER:
					printf("controller %i param %i \n",(int)ev->data.control.value,(int)ev->data.control.param);
					fflush(stdout);
					if ((int)ev->data.control.param==20&&(int)ev->data.control.value==127) {
						next();
					} else if ((int)ev->data.control.param==21&&(int)ev->data.control.value==127) {
						variation();
					} else if ((int)ev->data.control.param==22&&(int)ev->data.control.value==127) {
						//musicsystem->promptCue(MUSICCUE_JARSE_KORPIKUUSEN);
					} else if ((int)ev->data.control.param==23&&(int)ev->data.control.value==127) {
						//musicsystem->setParameterValue(MUSICPARAM_JARSE_NEXT,1);
						//musicsystem->setParameterValue(MUSICPARAM_JARSE_VARIATION,1);
					} 
					break; 
    			}
		}
		snd_seq_free_event(ev);
  	} while (snd_seq_event_input_pending(seq_handle, 0) > 0);
}


int main(int argc, char *argv[]) {
    
  FMOD::System      *system;
  FMOD::Event       *event;;
  FMOD_RESULT        result;
  int                key;

  printf("======================================================================\n");
  printf("Hurja. Experimental sampler. Copyright (c) Jari Suominen 2010-2011.\n");
  printf("======================================================================\n");

  ERRCHECK(result = FMOD::EventSystem_Create(&eventsystem));
  eventsystem->getSystemObject(&system);
  ERRCHECK(result = system->setOutput(FMOD_OUTPUTTYPE_ALSA));
  ERRCHECK(result = system->setSpeakerMode(FMOD_SPEAKERMODE_5POINT1)); 
  ERRCHECK(result = eventsystem->init(64, FMOD_INIT_NORMAL, 0, FMOD_EVENT_INIT_NORMAL));
  ERRCHECK(result = eventsystem->getMusicSystem(&musicsystem));
  //ERRCHECK(result = eventsystem->setMediaPath("./banks/"));
  ERRCHECK(result = eventsystem->setMediaPath(argv[1]));
  ERRCHECK(result = eventsystem->load("jarse.fev", 0, 0));
    
  int npfd;
  struct pollfd *pfd;

  seq_handle = open_seq();
  npfd = snd_seq_poll_descriptors_count(seq_handle, POLLIN);
  pfd = (struct pollfd *)alloca(npfd * sizeof(struct pollfd));
  snd_seq_poll_descriptors(seq_handle, pfd, npfd, POLLIN);

  printf("======================================================================\n");
  printf("Press ESC   to quit\n");
  printf("======================================================================\n");

  musicsystem->setParameterValue(MUSICPARAM_JARSE_NEXT,1);
  musicsystem->setParameterValue(MUSICPARAM_JARSE_VARIATION,1);
  musicsystem->setCallback(resetParams,0);

  char b[2];
  key = 0;

  do { // MAIN LOOP
        	
    if (poll(pfd,npfd, 100) > 0) {
      midi_action(seq_handle);
    }

    usleep (10);
	
    ERRCHECK(result = eventsystem->update());
    fflush(stdout);

  } while (1==1);

  ERRCHECK(result = eventsystem->release());
  return 0;
}
